unit asfClasses;

interface

uses
  SysUtils, Classes, Forms, Windows, DB, asfModel, asfData, Contnrs;

const
  { Log message types }
  lmLogIn             = 1;
  lmLogOut            = 2;
  lmUserAccountChange = 3;
  lmUserAccountAdd    = 4;
  lmUserAccessChange  = 5;
  lmUserOperation     = 100;

type
  TAppSecurityState = (ssInActive, ssLoggedIn, ssLoggedOut);
  TAppSecurityOption = (soAutoGenerateDatabase, soAutoLogin,
    soAutoLogout, soTerminateAppOnLoginFail, soAllowSelectUserInLogin);
  TAppSecurityOptions = set of TAppSecurityOption;

  TLogOptions = class(TPersistent)
  private
    FLoginMessage: String;
    FLogoutMessage: String;
    FRegisterLogin: Boolean;
    FRegisterLogout: Boolean;
    FUserAccountChangeMessage: String;
    FRegisterUserAccountChange: Boolean;
    FRegisterUserAccountAdd: Boolean;
    FUserAccountAddMessage: String;
    FRegisterUserAccessChange: Boolean;
    FUserAccessChangeMessage: String;
    FAllowDeleteRecords: Boolean;
    FUserOperationMessageType: Integer;
  public
    constructor Create;
  published
    property LoginMessage: String read FLoginMessage write FLoginMessage;
    property LogoutMessage: String read FLogoutMessage write FLogoutMessage;
    property UserAccountAddMessage: String read FUserAccountAddMessage write
      FUserAccountAddMessage;
    property UserAccountChangeMessage: String read FUserAccountChangeMessage write
      FUserAccountChangeMessage;
    property UserAccessChangeMessage: String read FUserAccessChangeMessage write
      FUserAccessChangeMessage;

    property RegisterLogin: Boolean read FRegisterLogin write FRegisterLogin
      default True;
    property RegisterLogout: Boolean read FRegisterLogout write FRegisterLogout
      default True;
    property RegisterUserAccountAdd: Boolean read FRegisterUserAccountAdd write
      FRegisterUserAccountAdd default True;
    property RegisterUserAccountChange: Boolean read FRegisterUserAccountChange
      write FRegisterUserAccountChange default True;
    property RegisterUserAccessChange: Boolean read FRegisterUserAccessChange write
      FRegisterUserAccessChange default True;

    property AllowDeleteRecords: Boolean read FAllowDeleteRecords write
      FAllowDeleteRecords default True;

    property UserOperationMessageType: Integer read FUserOperationMessageType write
      FUserOperationMessageType default lmUserOperation;
  end;

  TApplicationUser = class(TObject)
  private
    FID: Integer;
    FLogin: String;
    FFullName: String;
  public
    constructor Create(AID: Integer; const ALogin, AFullName: String);
    property ID: Integer read FID;
    property Login: String read FLogin;
    property FullName: String read FFullName;
  end;

  TItemIDList = class(TList)
  private
    function GetItems(Index: Integer): Integer;
  public
    procedure Add(ItemID: Integer);
    function IndexOf(ItemID: Integer): Integer;
    property Items[Index: Integer]: Integer read GetItems; default;
  end;

  TShowLogRecordDataEvent = procedure(Sender: TObject; ALogRecordID,
    ALogRecordType: Integer) of object;

  TAppSecurity = class(TComponent)
  private
    FActive: Boolean;
    FDataProvider: TAppSecurityDataProvider;
    FModel: TAppSecurityModel;
    FState: TAppSecurityState;
    FCurrentUser: TApplicationUser;
    FFormControllers: TComponentList;
    FLogOptions: TLogOptions;
    FOnLogin: TNotifyEvent;
    FOnLogout: TNotifyEvent;
    FOptions: TAppSecurityOptions;
    FLoginNumberOfTries: Integer;
    FMinPasswordLength: Integer;
    FOnCreateDefaultUsers: TNotifyEvent;
    FDefultUserName: String;
    FDefultUserLogin: String;
    FDefultUserPassword: String;
    FLoginDialogHint: String;
    FOnShowLogRecordData: TShowLogRecordDataEvent;
    procedure SetActive(Value: Boolean);
    procedure SetDataProvider(Value: TAppSecurityDataProvider);
    procedure SetModel(Value: TAppSecurityModel);
    procedure CheckActive;
    function GetSecurityItemStates(ItemIDList: TItemIDList; AStates: TStrings):
      String;
    procedure CheckState(ARequiredState: TAppSecurityState; const AOperationName:
      string);
    procedure SetLogOptions(const Value: TLogOptions);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DoLogin; dynamic;
    procedure DoLogout; dynamic;
    procedure CreateDefaultUser; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddLogRecord(AMessageType: Integer; AMessage: String; AUserData:
      TObject = nil);
    procedure GenerateDatabase;
    function Login: Boolean;
    function Logout: Boolean;
    procedure SetupFormAccess(AForm: TComponent);
    function AddUser(AUserName, ALogin, APassword: String; ADefaultAccessKind:
      String = ''): Integer;
    function GetLastLogRecord: TDataSet;
    procedure RegisterUserOperation(AMessage: String);
    property Active: Boolean read FActive write SetActive;
    property State: TAppSecurityState read FState;
    property CurrentUser: TApplicationUser read FCurrentUser;
  published
    property Model: TAppSecurityModel read FModel write SetModel;
    property DataProvider: TAppSecurityDataProvider read FDataProvider write
      SetDataProvider;

    property LogOptions: TLogOptions read FLogOptions write SetLogOptions;
    property Options: TAppSecurityOptions read FOptions write FOptions default
      [soAutoGenerateDatabase, soAutoLogin,
       soTerminateAppOnLoginFail];
    property LoginNumberOfTries: Integer read FLoginNumberOfTries write
      FLoginNumberOfTries default 3;
    property LoginDialogHint: String read FLoginDialogHint write FLoginDialogHint;
    property MinPasswordLength: Integer read FMinPasswordLength write
      FMinPasswordLength default 6;
    property DefultUserName: String read FDefultUserName write FDefultUserName;
    property DefultUserLogin: String read FDefultUserLogin write FDefultUserLogin;
    property DefultUserPassword: String read FDefultUserPassword write
      FDefultUserPassword;

    property OnLogin: TNotifyEvent read FOnLogin write FOnLogin;
    property OnLogout: TNotifyEvent read FOnLogout write FOnLogout;
    property OnCreateDefaultUsers: TNotifyEvent read FOnCreateDefaultUsers write
      FOnCreateDefaultUsers;
    property OnShowLogRecordData: TShowLogRecordDataEvent read FOnShowLogRecordData
      write FOnShowLogRecordData;
  end;

  TAppSecurityFormControllerItem = class(TCollectionItem)
  private
    FComponentName: String;
    FFormName: String;
    FPropertyName: String;
    FAccessState: String;
    FModelItemID: Integer;
    FPropertyValue: Variant;
  public
    procedure Assign(Source: TPersistent); override;
  published
    property ModelItemID: Integer read FModelItemID write FModelItemID;
    property AccessState: String read FAccessState write FAccessState;
    property FormName: String read FFormName write FFormName;
    property ComponentName: String read FComponentName write FComponentName;
    property PropertyName: String read FPropertyName write FPropertyName;
    property PropertyValue: Variant read FPropertyValue write FPropertyValue;
  end;

  TAppSecurityFormController = class(TComponent)
  private
    FAppSecurity: TAppSecurity;
    FItems: TCollection;
    procedure SetAppSecurity(Value: TAppSecurity);
    procedure SetupItems(AForm: TComponent; Items: TList);
    function GetItemCount: Integer;
    function GetItems(Index: Integer): TAppSecurityFormControllerItem;
    procedure ReadItems(Reader: TReader);
    procedure WriteItems(Writer: TWriter);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DefineProperties(Filer: TFiler); override;
  public
    procedure SetupFormAccess(AForm: TComponent);
    {{
       
    }
    function AddItem: TAppSecurityFormControllerItem;
    procedure Assign(Source: TPersistent); override;
    procedure Clear;
    {{
           
    }
    procedure DeleteItem(AItem: TAppSecurityFormControllerItem);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure GetLinkedModelItems(AFormName: String; ItemIDList: TItemIDList);
    procedure GetItemsForStates(AFormName: String; AStates: TStrings; AItems:
      TList);
    property ItemCount: Integer read GetItemCount;
    property Items[Index: Integer]: TAppSecurityFormControllerItem read GetItems;
  published
    property AppSecurity: TAppSecurity read FAppSecurity write SetAppSecurity;
  end;

function DefaultAppSecurity: TAppSecurity;

implementation

uses asfUtils, asfLoginDlg;

var
  _DefaultAppSecurity: TAppSecurity = nil;

function DefaultAppSecurity: TAppSecurity;
begin
  Result := _DefaultAppSecurity
end;

function GetComputerName: String;
var
  Buffer: array [0..MAX_COMPUTERNAME_LENGTH] of Char;
  Size: Cardinal;
begin
  Windows.GetComputerName(Buffer, Size);
  Result := String(PChar(@Buffer))
end;

{ TLogOptions }

constructor TLogOptions.Create;
begin
  inherited Create;

  FRegisterLogin := True;
  FRegisterLogout := True;
  FRegisterUserAccountChange := True;
  FRegisterUserAccountAdd := True;
  FRegisterUserAccessChange := True;

  FLoginMessage := '   .';
  FLogoutMessage := '   .';
  FUserAccountAddMessage := '   : ';
  FUserAccountChangeMessage := '   : ';
  FUserAccessChangeMessage := '   : ';

  FAllowDeleteRecords := True;

  FUserOperationMessageType := lmUserOperation;
end;

{ TApplicationUser }

constructor TApplicationUser.Create(AID: Integer; const ALogin, AFullName:
  String);
begin
  inherited Create;
  FID := AID;
  FLogin := ALogin;
  FFullName := AFullName;
end;

{ TItemIDList }

procedure TItemIDList.Add(ItemID: Integer);
begin
  inherited Add(Pointer(ItemID))
end;

function TItemIDList.GetItems(Index: Integer): Integer;
begin
  Result := Integer(inherited Items[Index]);
end;

function TItemIDList.IndexOf(ItemID: Integer): Integer;
begin
  Result := inherited IndexOf(Pointer(ItemID));
end;

{ TAppSecurity }

constructor TAppSecurity.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if Not Assigned(_DefaultAppSecurity) then _DefaultAppSecurity := Self;
  FFormControllers := TComponentList.Create(False);
  FLogOptions := TLogOptions.Create;
  FState := ssInActive;
  FOptions := [soAutoGenerateDatabase, soAutoLogin, soTerminateAppOnLoginFail];
  FLoginNumberOfTries := 3;
  FMinPasswordLength := 6;
  FDefultUserName := '';
  FDefultUserLogin := 'user';
end;

destructor TAppSecurity.Destroy;
begin
  Active := False;

  FreeAndNil(FCurrentUser);
  FreeAndNil(FLogOptions);
  FreeAndNil(FFormControllers);
  if _DefaultAppSecurity = Self then _DefaultAppSecurity := nil;
  inherited Destroy;
end;

procedure TAppSecurity.SetModel(Value: TAppSecurityModel);
begin
  if FModel <> Value then
  begin
    if Value = nil then Active := False;
    FModel := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TAppSecurity.SetDataProvider(Value: TAppSecurityDataProvider);
begin
  if FDataProvider <> Value then
  begin
    if Value = nil then Active := False;
    FDataProvider := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

{{
    published .
}
procedure TAppSecurity.SetLogOptions(const Value: TLogOptions);
begin
end;

procedure TAppSecurity.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Model) and (Operation = opRemove) then Model := nil else
  if (AComponent = DataProvider) and (Operation = opRemove) then DataProvider := nil;
end;

procedure TAppSecurity.CheckState(ARequiredState: TAppSecurityState; const
  AOperationName: string);
begin
  if FState <> ARequiredState then
    raise Exception.Create('       ' + AOperationName);
end;

procedure TAppSecurity.CheckActive;
begin
  if Not Active then raise Exception.Create('   : ' + Name);
end;

procedure TAppSecurity.DoLogin;
begin
  if Assigned(FOnLogin) then FOnLogin(Self);
end;

procedure TAppSecurity.DoLogout;
begin
  if Assigned(FOnLogout) then FOnLogout(Self);
end;

{***************************************************************************** }
{                                                              }
{ **************************************************************************** }

procedure TAppSecurity.SetActive(Value: Boolean);
begin
  if FActive <> Value then
  begin
    if Value then
    begin
      if Not Assigned(Model) then
        raise Exception.Create('    ' + Name);
      if Not Assigned(DataProvider) then
        raise Exception.Create('     ' + Name);
    end;

    if Not Value and (soAutoLogout in Options) then Logout;

    FActive := Value;
    if Active then FState := ssLoggedOut else FState := ssInActive;

    if Value then
    begin
      if soAutoGenerateDatabase in Options then GenerateDatabase;
      if soAutoLogin in Options then
      begin
        if Not Login and (soTerminateAppOnLoginFail in Options) then
        begin
          Application.MessageBox('    .'#13#10+
            '  .', '', MB_ICONERROR or MB_OK);
          Application.Terminate;
        end;
      end;
    end;
  end;
end;

function TAppSecurity.Login: Boolean;
var
  AppUser: TApplicationUser;
begin
  Result := False;
  CheckState(ssLoggedOut, 'LogIn');
  if ShowLoginDialog(Self, AppUser, LoginDialogHint) = lrSuccess then
  begin
    FCurrentUser := AppUser;
    FState := ssLoggedIn;

    if LogOptions.RegisterLogin then AddLogRecord(lmLogIn, LogOptions.LogInMessage);
    DoLogin;

    Result := True;
  end;
end;

function TAppSecurity.Logout: Boolean;
begin
  if State = ssLoggedIn then
  begin
    DoLogout;
    if LogOptions.RegisterLogOut then AddLogRecord(lmLogOut, LogOptions.LogOutMessage);

    FreeAndNil(FCurrentUser);
    FState := ssLoggedOut;
    Result := True;
  end else
    Result := False
end;

procedure TAppSecurity.AddLogRecord(AMessageType: Integer; AMessage: String;
  AUserData: TObject = nil);
begin
  if State = ssLoggedIn then
    DataProvider.AddLogRecord(Now, CurrentUser.ID, CurrentUser.FullName, GetComputerName,
      AMessageType, AMessage, AUserData);
end;

procedure TAppSecurity.RegisterUserOperation(AMessage: String);
begin
  AddLogRecord(LogOptions.UserOperationMessageType, AMessage)
end;

function TAppSecurity.GetLastLogRecord: TDataSet;
var
  LogRecordID: Integer;
begin
  CheckState(ssLoggedIn, 'AddLogMessage');
  LogRecordID := DataProvider.GetLastLogRecordID;
  Result := DataProvider.GetLogRecord(LogRecordID);
end;

procedure TAppSecurity.GenerateDatabase;
begin
  CheckActive;
  if DataProvider.GenerateDatabase
  {      ,   /   }
    then CreateDefaultUser;
end;

procedure TAppSecurity.CreateDefaultUser;
begin
  if Assigned(FOnCreateDefaultUsers) then FOnCreateDefaultUsers(Self)
  else
    {            }
    AddUser(DefultUserName, DefultUserLogin, DefultUserPassword, Model.EnabledState);
end;

procedure TAppSecurity.SetupFormAccess(AForm: TComponent);
var
  FC: Pointer;
begin
  if Assigned(Self) and Active then
  begin
    CheckState(ssLoggedIn, 'SetupFormAccess');
    for FC in FFormControllers do
      TAppSecurityFormController(FC).SetupFormAccess(AForm);
  end;
end;

{***************************************************************************** }
{                                                                              }
{ **************************************************************************** }

{{
        
,    ItemIDList.
}
function TAppSecurity.GetSecurityItemStates(ItemIDList: TItemIDList; AStates:
  TStrings): String;
var
  AccessKind: String;
  I: Integer;
begin
  for I := 0 to ItemIDList.Count - 1 do
    if DataProvider.GetObjectRights(CurrentUser.ID, UserObjectType, ItemIDList[I], AccessKind)
      then AStates.AddObject(AccessKind, TObject(ItemIDList[I]));
end;

function TAppSecurity.AddUser(AUserName, ALogin, APassword: String;
  ADefaultAccessKind: String = ''): Integer;
var
  ModelItem: TAppSecurityModelItem;
begin
  CheckActive;
  Result := DataProvider.AddUser(AUserName);
  DataProvider.SetUserInfo(Result, AUserName, ALogin, APassword, 0);

  if ADefaultAccessKind <> '' then
  begin
    {        }
    DataProvider.ClearObjectRights(Result, UserObjectType);
    for ModelItem in Model do
      if ModelItem.HasState(ADefaultAccessKind) then
        DataProvider.SetObjectRights(Result, UserObjectType, ModelItem.ID, ADefaultAccessKind);
  end;
end;

{ TAppSecurityFormController }

constructor TAppSecurityFormController.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TCollection.Create(TAppSecurityFormControllerItem);
end;

destructor TAppSecurityFormController.Destroy;
begin
  FreeAndNil(FItems);
  inherited Destroy;
end;

procedure TAppSecurityFormController.SetAppSecurity(Value: TAppSecurity);
begin
  if FAppSecurity <> Value then
  begin
    if Assigned(FAppSecurity) then FAppSecurity.FFormControllers.Remove(Self);

    FAppSecurity := Value;
    if Value <> nil then Value.FreeNotification(Self);

    if Assigned(FAppSecurity) then FAppSecurity.FFormControllers.Add(Self);
  end;
end;

procedure TAppSecurityFormController.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = AppSecurity) and (Operation = opRemove) then AppSecurity := nil
end;

{***************************************************************************** }
{                                                        }
{ **************************************************************************** }

function TAppSecurityFormController.GetItemCount: Integer;
begin
  Result := FItems.Count
end;

function TAppSecurityFormController.GetItems(Index: Integer):
  TAppSecurityFormControllerItem;
begin
  Result := TAppSecurityFormControllerItem(FItems.Items[Index])
end;

procedure TAppSecurityFormController.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Items', ReadItems, WriteItems, FItems.Count > 0);
end;

procedure TAppSecurityFormController.ReadItems(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FItems)
end;

procedure TAppSecurityFormController.WriteItems(Writer: TWriter);
begin
  Writer.WriteCollection(FItems)
end;

function TAppSecurityFormController.AddItem: TAppSecurityFormControllerItem;
begin
  Result := TAppSecurityFormControllerItem(FItems.Add);
end;

procedure TAppSecurityFormController.DeleteItem(AItem:
  TAppSecurityFormControllerItem);
begin
  AItem.Collection.Delete(AItem.Index);
end;

procedure TAppSecurityFormController.Clear;
begin
  FItems.Clear;
end;

{{
    Source.
}
procedure TAppSecurityFormController.Assign(Source: TPersistent);
begin
  if Source is TAppSecurityFormController then
  begin
    with TAppSecurityFormController(Source) do
      Self.FItems.Assign(FItems);
  end else
    inherited Assign(Source);
end;

{***************************************************************************** }
{                                                        }
{ **************************************************************************** }

{{
        ,
      .
}
procedure TAppSecurityFormController.SetupFormAccess(AForm: TComponent);
var
  ItemIDList: TItemIDList;
  ItemStates: TStrings;
begin
  ItemIDList := TItemIDList.Create;
  ItemStates := TStringList.Create;
  try

    {    ,      }
    GetLinkedModelItems(AForm.ClassName, ItemIDList);

    {     ,    }
    AppSecurity.GetSecurityItemStates(ItemIDList, ItemStates);

    {    TAppSecurityFormControllerItem,     }
    GetItemsForStates(AForm.ClassName, ItemStates, ItemIDList);

    SetupItems(AForm, ItemIDList);

  finally
    ItemStates.Free;
    ItemIDList.Free;
  end;
end;

{{
   ,     
}
procedure TAppSecurityFormController.GetLinkedModelItems(AFormName: String;
  ItemIDList: TItemIDList);
var
  I: Integer;
begin
  ItemIDList.Clear;
  for I := 0 to ItemCount - 1 do
    if (Items[I].FormName = AFormName) and (ItemIDList.IndexOf(Items[I].ModelItemID) < 0)
      then ItemIDList.Add(Items[I].ModelItemID);
end;

procedure TAppSecurityFormController.GetItemsForStates(AFormName: String;
  AStates: TStrings; AItems: TList);

  procedure SelectItems(AModelItemID: Integer; AAccessState: String);
  var
    I: Integer;
  begin
    for I := 0 to ItemCount - 1 do
      if (Items[I].FormName = AFormName) and (Items[I].ModelItemID = AModelItemID) and
        (Items[I].AccessState = AAccessState) then AItems.Add(Items[I]);
  end;

var
  I: Integer;
begin
  AItems.Clear;
  for I := 0 to AStates.Count - 1 do
    SelectItems(Integer(AStates.Objects[I]), AStates[I])
end;

{{
        Items 
TAppSecurityFormControllerItem.
}
procedure TAppSecurityFormController.SetupItems(AForm: TComponent; Items:
  TList);

  procedure SetupItem(Item: TAppSecurityFormControllerItem);
  begin
    {   -    }
    SetObjectPropertyValue(AForm.FindComponent(Item.ComponentName), Item.PropertyName, Item.PropertyValue)
  end;

var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    SetupItem(TAppSecurityFormControllerItem(Items[I]))
end;

{ TAppSecurityFormControllerItem }

{{
   Source   .
}
procedure TAppSecurityFormControllerItem.Assign(Source: TPersistent);
begin
  if Source is TAppSecurityFormControllerItem then
  begin
    with TAppSecurityFormControllerItem(Source) do
    begin
      {    }
      Self.ModelItemID := ModelItemID;
      Self.AccessState := AccessState;
      Self.FormName := FormName;
      Self.ComponentName := ComponentName;
      Self.PropertyName := PropertyName;
      Self.PropertyValue := PropertyValue;
    end;
  end else
    inherited Assign(Source);
end;

end.
